/*
 * ${license.title}
 * Copyright (C) 2022 ${license.owner}
 * ${license.mailto}
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "rollball.hpp"
#include "fggl/gfx/camera.hpp"
#include "fggl/gfx/phong.hpp"

#include "fggl/input/camera_input.hpp"
#include "fggl/util/service.hpp"
#include "fggl/ecs3/prototype/loader.hpp"
#include "fggl/debug/draw.hpp"

#include <array>

struct Prefabs {
	fggl::ecs3::entity_t collectable;
	fggl::ecs3::entity_t player;
};

static void setup_prefabs(fggl::ecs3::World& world, Prefabs& prefabs) {

	auto storage = fggl::util::ServiceLocator::instance().get<fggl::data::Storage>();
	fggl::ecs3::load_prototype_file(world, *storage, "rollball.yml");

	{
		// player (cube because my sphere function doesn't exist yet
		prefabs.player = world.findPrototype("player");
		world.add<fggl::phys::Dynamics>(prefabs.player);
	}

	{
		// collectable
		prefabs.collectable = world.findPrototype("collectable");

		// we need both of these for callbacks to trigger.
		world.add<fggl::phys::CollisionCallbacks>(prefabs.collectable);
		world.add<fggl::phys::CollisionCache>(prefabs.collectable);
	}

}

static void setup_camera(fggl::ecs3::World& world) {
	auto prototype = world.create(false);

	// setup camera position/transform
	auto* transform = world.add<fggl::math::Transform>(prototype);
	if ( transform != nullptr) {
		transform->origin(glm::vec3(10.0f, 3.0f, 10.0f));
	}

	// setup camera components
	world.add<fggl::gfx::Camera>(prototype);
}

static fggl::ecs3::entity_t setup_environment(fggl::ecs3::World& world, const fggl::math::vec2& size, demo::RollState& state) {
	{
		auto northWall = world.createFromPrototype("wallX");
		auto* transform = world.get<fggl::math::Transform>(northWall);
		transform->origin({size.x/2, 0.0F, 0.0F});
	}

	{
		auto southWall = world.createFromPrototype("wallX");
		auto* transform = world.get<fggl::math::Transform>(southWall);
		transform->origin({-size.x/2, 0.0F, 0.0F});
	}

	{
		auto westWall = world.createFromPrototype("wallZ");
		auto* transform = world.get<fggl::math::Transform>(westWall);
		transform->origin({0.0F, 0.0F, -size.y/2});
	}

	{
		auto eastWall = world.createFromPrototype("wallZ");
		auto* transform = world.get<fggl::math::Transform>(eastWall);
		transform->origin({0.0F, 0.0F, size.y/2});
	}

	{
		auto floor = world.createFromPrototype("floor");
		auto *transform = world.get<fggl::math::Transform>(floor);
		transform->origin({0.0F, -2.5F, 0.0F});
	}

	{
		// player just starts off as the prefab dictates
		state.player = world.createFromPrototype("player");
	}

	{
		// collectables
		std::array<fggl::math::vec3, 3> collectPos {{
			{-5.0f, -0.5f, 12.0f},
			{15.0f, -0.5f, 0.5f},
			{6.0f, -0.5f, -15.0f}
		}};

		// build the collectables
		int collectCount = 0;
		for (auto& pos : collectPos) {
			auto collectable = world.createFromPrototype("collectable");
			auto* transform = world.get<fggl::math::Transform>(collectable);
			transform->origin(pos);

			state.collectables[collectCount++] = collectable;
		}
	}

	// ensure the state is clean
	state.closestPickup = fggl::ecs::NULL_ENTITY;
	state.mode = demo::DebugMode::NORMAL;
	state.time = 0.0F;

	return state.player;
}


namespace demo {

	constexpr fggl::math::vec2 WORLD_SIZE{40.0F, 40.0F };
	constexpr fggl::math::vec3 COLLECTABLE_ROTATION{15, 30, 45};
	constexpr fggl::math::vec3 COLOUR_WHITE{1.0F, 1.0F, 1.0F};
	constexpr fggl::math::vec3 COLOUR_BLUE{0.0F, 0.0F, 1.0F};
	constexpr float MOVE_FORCE{3.0F};

	RollBall::RollBall(fggl::App &app) : Game(app) {
	}

	void RollBall::activate() {
		Game::activate();

		Prefabs prefabs{};
		setup_prefabs(world(), prefabs);

		// collectable callbacks
		auto* collectableCallbacks = world().get<fggl::phys::CollisionCallbacks>(prefabs.collectable);
		collectableCallbacks->onEnter = [this](auto ourEntity, auto theirEntity) {
			if ( theirEntity == state.player) {
				//if ( ourEntity == state.closestPickup ) {
				//	// we're the closest pickup and we're about to get killed, so need to not be.
				//	state.closestPickup = fggl::ecs::NULL_ENTITY;
				//}
				this->world().destroy(ourEntity);
			}
		};

		// actual scene objects
		setup_camera(world());

		// create a 20x20 grid
		setup_environment(world(), WORLD_SIZE, state);
	}

	fggl::math::vec3 calc_move_vector(const fggl::input::Input& input) {
		constexpr fggl::math::vec3 forward = fggl::math::FORWARD;
		constexpr fggl::math::vec3 right = fggl::math::RIGHT;
		fggl::math::vec3 force = fggl::math::VEC3_ZERO;

		if (input.keyboard.down(glfwGetKeyScancode(GLFW_KEY_W))) {
			force += forward;
		}

		if (input.keyboard.down(glfwGetKeyScancode(GLFW_KEY_S))) {
			force -= forward;
		}

		if (input.keyboard.down(glfwGetKeyScancode(GLFW_KEY_A))) {
			force -= right;
		}

		if (input.keyboard.down(glfwGetKeyScancode(GLFW_KEY_D))) {
			force += right;
		}

		return force;
	}

	void RollBall::update() {
		Game::update();

		const float deltaTime = 1 / 60.0F;

		auto& input = this->input();

			if ( state.player != fggl::ecs3::NULL_ENTITY ) {
				auto &world = this->world();

				// mode selection
				if (input.keyboard.pressed(glfwGetKeyScancode(GLFW_KEY_SPACE))) {
					state.mode = (DebugMode)( ((int)state.mode + 1) % 3 );
				}

				// setup dynamic force
				auto force = calc_move_vector(input);
				if ( force != fggl::math::VEC3_ZERO ) {
					force = glm::normalize(force) * MOVE_FORCE;
					auto *dynamics = world.get<fggl::phys::Dynamics>(state.player);
					dynamics->force += force;
				}

				// track player position with camera
				{
					auto cameras = world.findMatching<fggl::gfx::Camera>();
					fggl::ecs3::entity_t cam = cameras[0];
					auto *camComp = world.get<fggl::gfx::Camera>(cam);
					camComp->target = world.get<fggl::math::Transform>(state.player)->origin();

					auto *camTransform = world.get<fggl::math::Transform>(cam);
					camTransform->origin( camComp->target + cameraOffset );
				}

				// distance mode
				if ( state.mode == DebugMode::DISTANCE ) {
					closestPickup(world);
				} else {
					if ( state.closestPickup != fggl::ecs::NULL_ENTITY ) {
						if ( world.alive(state.closestPickup) ){
							auto *renderer = world.tryGet<fggl::gfx::PhongMaterial>(state.closestPickup);
							if (renderer != nullptr) {
								renderer->diffuse = COLOUR_WHITE;
							}
						}

						state.closestPickup = fggl::ecs::NULL_ENTITY;
					}
				}

				// make the cubes spin
				spinCubes(world, deltaTime);
				state.time += deltaTime;
		}
	}

	void RollBall::closestPickup(fggl::ecs3::World &world) {
		float closestDistance = FLT_MAX;
		fggl::ecs::entity_t closestEntity = fggl::ecs::NULL_ENTITY;
		auto playerPos = world.get<fggl::math::Transform>(state.player)->origin();

		for ( const auto& pickup : state.collectables ) {
			if ( !world.alive(pickup) ) {
				continue;
			}

			auto* transform = world.get<fggl::math::Transform>(pickup);
			auto distance = glm::distance(transform->origin(), playerPos);
			if ( distance < closestDistance) {
				closestDistance = distance;
				closestEntity = pickup;
			}
		}

		if ( state.closestPickup != closestEntity ) {
			if ( world.alive(state.closestPickup) ){
				// deal with the previous closest pickup
				auto *renderer = world.tryGet<fggl::gfx::PhongMaterial>(state.closestPickup);
				if (renderer != nullptr) {
					renderer->diffuse = COLOUR_WHITE;
				}
			}

			// now, deal with the new closest
			state.closestPickup = closestEntity;
			if (closestEntity != fggl::ecs::NULL_ENTITY) {
				auto *renderer = world.tryGet<fggl::gfx::PhongMaterial>(state.closestPickup);
				if (renderer != nullptr) {
					renderer->diffuse = COLOUR_BLUE;
				}
			}
		}

		// closest pickup should face the player
		if ( state.closestPickup != fggl::ecs::NULL_ENTITY) {
			auto* transform = world.get<fggl::math::Transform>(state.closestPickup);
			auto* playerTransform = world.get<fggl::math::Transform>( state.player );
			transform->lookAt(playerTransform->origin());

			// lazy, so using the debug draw line for this bit
			dd::line( &playerTransform->origin()[0], &transform->origin()[0], dd::colors::White );
		}

	}

	void RollBall::spinCubes(fggl::ecs3::World& world, float deltaTime) {
		// rotation
		for ( const auto& entity : state.collectables ) {
			if ( !world.alive(entity) || entity == state.closestPickup ) {
				continue;
			}

			// rotate the cubes
			auto* transform = world.get<fggl::math::Transform>(entity);
			transform->rotateEuler(  COLLECTABLE_ROTATION * deltaTime );
		}

	}

	void RollBall::render(fggl::gfx::Graphics &gfx) {
		Game::render(gfx);
	}

} // namespace demo

