#include "fggl/ecs3/prototype/loader.hpp"

namespace fggl::ecs3 {

	void load_prototype_node( ecs3::World& world, const YAML::Node& node) {
		auto prototype = world.create(true);
		auto& typeReg = world.types();

		// store the prototype's name for future use
		auto* nameComp = world.get<ecs3::EntityMeta>( prototype );
		nameComp->typeName = node["name"].as<std::string>();

		// grab the components and iterate them
		if ( node["components"] ) {
			world.createFromSpec( prototype, node["components"]);
		}
	}

	void load_prototype_file( ecs3::World& world, data::Storage& storage, const std::string& name ) {
		auto path = storage.resolvePath( data::Data, name );

		auto root = YAML::LoadFile( path.string().c_str() );
		for (const auto& node : root["prefabs"] ) {
			load_prototype_node(world, node);
		}
	}

}

namespace fggl::ecs {

	static void process_mesh(data::Mesh& out, const YAML::Node& shape) {
		auto transform = data::OFFSET_NONE;
		if ( shape["offset"] ) {
			auto scaleFactor = shape["offset"].as<math::vec3>();
			transform = glm::translate(transform,  scaleFactor);
		}

		if ( shape["scale"] ) {
			auto scaleFactor = shape["scale"].as<math::vec3>();
			transform = glm::scale(transform,  scaleFactor);
		}

		// now the shape itself
		auto shapeType = shape["type"].as<std::string>();
		if ( shapeType == SHAPE_BOX_VALUE ) {
			data::make_cube(out, transform);
		} else if ( shapeType == SHAPE_SPHERE_VALUE ) {
			int stacks = shape["stacks"].as<int>(DEFAULT_STACKS);
			int slices = shape["slices"].as<int>(DEFAULT_SLICES);
			data::make_sphere(out, transform, stacks, slices);
		}
	}

	template<>
	bool restore_config(math::Transform* comp, const YAML::Node& node) {
		comp->origin(node["origin"].as<math::vec3>( math::VEC3_ZERO ));
		comp->euler(node["rotation"].as<math::vec3>(math::VEC3_ZERO));
		comp->scale(node["scale"].as<math::vec3>(math::VEC3_ONES));
		return true;
	}

	template<>
	bool restore_config(gfx::PhongMaterial* comp, const YAML::Node& node) {
		comp->diffuse = node["diffuse"].as<math::vec3>( gfx::DEFAULT_DIFFUSE );
		comp->ambient = node["ambient"].as<math::vec3>( gfx::DEFAULT_AMBIENT );
		comp->specular = node["specular"].as<math::vec3>( gfx::DEFAULT_SPECULAR );
		comp->shininess = node["shininess"].as<float>( gfx::DEFAULT_SHININESS );
		return true;
	}

	template<>
	bool restore_config(data::StaticMesh* meshComp, const YAML::Node& node) {
		if ( !node["pipeline"] ) {
			return false;
		}

		meshComp->pipeline = node["pipeline"].as<std::string>();

		if ( node["shape"] ) {
			data::Mesh mesh;
			if (node["shape"].IsSequence()) {
				// is a composite shape
				for (const auto &shapeNode : node["shape"]) {
					process_mesh(mesh, shapeNode);
				}
			} else {
				// is a basic shape
				process_mesh(mesh, node["shape"]);
			}

			// optimise and load
			mesh.removeDups();
			meshComp->mesh = mesh;
		} else {
			return false;
		}

		return true;
	}

	template<>
	bool restore_config(phys::RigidBody* body, const YAML::Node& node) {
		body->type = node["type"].as<fggl::phys::BodyType>(fggl::phys::BodyType::DYNAMIC);
		if ( body->type == fggl::phys::BodyType::STATIC) {
			body->mass = phys::MASS_STATIC;
		} else {
			body->mass = node["mass"].as<float>( phys::MASS_DEFAULT );
		}

		// shape detection
		if ( node["shape"] ) {
			auto type = node["shape"]["type"].as< std::string >();

			if ( type == SHAPE_BOX_VALUE ) {
				// assume unit box if extents are missing
				auto extents = node["shape"]["extents"].as<math::vec3>(phys::UNIT_EXTENTS);
				body->shape = new phys::Box(extents);
			} else if ( type == SHAPE_SPHERE_VALUE ) {
				auto radius = node["shape"]["radius"].as<float>(0.5F);
				body->shape = new phys::Sphere(radius);
			} else {
				return false;
			}
		}

		return true;
	}

	template<>
	bool restore_config(phys::CollisionCallbacks* callbacks, const YAML::Node& node) {
		// TODO implement callback scripting
		return true;
	}

	template<>
	bool restore_config(phys::CollisionCache* cache, const YAML::Node& node){
		// probably nothing needed here. Ideally, this should be automatically added when callbacks are.
		return true;
	}
}
