/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

#include <fggl/util/service.hpp>
#include "fggl/debug/logging.hpp"

#include "fggl/gfx/ogl/common.hpp"
#include "fggl/gfx/window.hpp"

#include <fggl/gfx/camera.hpp>
#include <fggl/gfx/ogl/renderer.hpp>
#include <fggl/gfx/paint.hpp>
#include "fggl/data/model.hpp"
#include "fggl/gfx/ogl4/fallback.hpp"

#include <glm/ext/matrix_clip_space.hpp>
#include <glm/ext/matrix_transform.hpp>
#include <glm/glm.hpp>
#include <glm/gtc/type_ptr.hpp>

#include <memory>

extern "C" {

/**
 * Convert an OpenGL source enum to a string.
 *
 * list of sources taken from table 20.1, GL Spec 4.5
 * @param source
 * @return string representing the source
 */
constexpr auto fggl_ogl_source(GLenum source) -> const char * {
	switch (source) {
	case GL_DEBUG_SOURCE_API: return "GL";
	case GL_DEBUG_SOURCE_SHADER_COMPILER: return "GLSL compiler";
	case GL_DEBUG_SOURCE_WINDOW_SYSTEM: return "Windowing System";
	case GL_DEBUG_SOURCE_THIRD_PARTY: return "Third Party";
	case GL_DEBUG_SOURCE_APPLICATION: return "Application";
	default:
	case GL_DEBUG_SOURCE_OTHER: return "Other";
	}
	assert(false);
	return "unknown";
}

/**
 * Convert an OpenGL type enum to a string.
 *
 * list of sources taken from table 20.2, GL Spec 4.5
 * @param type
 * @return string representing the type
 */
constexpr auto static fggl_ogl_type(GLenum type) -> const char * {
	switch (type) {
	case GL_DEBUG_TYPE_ERROR: return "Error";
		break;
	case GL_DEBUG_TYPE_DEPRECATED_BEHAVIOR: return "Deprecated Behaviour";
		break;
	case GL_DEBUG_TYPE_UNDEFINED_BEHAVIOR: return "Undefined Behaviour";
		break;
	case GL_DEBUG_TYPE_PERFORMANCE: return "Performance";
		break;
	case GL_DEBUG_TYPE_PORTABILITY: return "Portability";
		break;
	case GL_DEBUG_TYPE_MARKER: return "Marker";
		break;
	case GL_DEBUG_TYPE_PUSH_GROUP: return "Push Group";
		break;
	case GL_DEBUG_TYPE_POP_GROUP: return "Pop Group";
		break;
	default:
	case GL_DEBUG_TYPE_OTHER: return "Other";
		break;
	}
	assert(false);
	return "unknown";
}

#ifndef _MSC_VER
	#pragma clang diagnostic push
	#pragma ide diagnostic ignored "bugprone-easily-swappable-parameters"
#endif

constexpr const char* OGL_LOG_FMT {"[GL] {}, {}: [{}]: {}"};

void static GLAPIENTRY fggl_ogl_to_spdlog(GLenum source, GLenum type, unsigned int msgId, GLenum severity, GLsizei  /*length*/,
							   const char *message, const void * /*userParam*/) {

	const auto *const sourceStr = fggl_ogl_source(source);
	const auto *const typeStr = fggl_ogl_type(type);

	// table 20.3, GL spec 4.5
	switch (severity) {
		case GL_DEBUG_SEVERITY_HIGH: fggl::debug::error(OGL_LOG_FMT, sourceStr, typeStr, msgId, message);
			break;
		default:
		case GL_DEBUG_SEVERITY_MEDIUM: fggl::debug::warning(OGL_LOG_FMT, sourceStr, typeStr, msgId, message);
			break;
		case GL_DEBUG_SEVERITY_LOW: fggl::debug::info(OGL_LOG_FMT, sourceStr, typeStr, msgId, message);
			break;
		case GL_DEBUG_SEVERITY_NOTIFICATION: fggl::debug::debug(OGL_LOG_FMT, sourceStr, typeStr, msgId, message);
			break;
	}
}

#ifndef _MSC_VER
	#pragma clang diagnostic pop
#endif
}

/**
 * Future optimisations:
 *   recommended approach is to group stuff in to as few vao as possible - this
 * will do one vao per mesh, aka bad. Add support for instanced rendering
 * (particles) Look at packing vertex data in better ways (with profiling)
 *   Support shader specialisation (ie, dynamic/streamed data)
 *   Follow best recommendations for Vertex attributes (ie normals not using
 * floats, bytes for colour)
 *
 * Future features:
 *   Add support for models with weights (for animations)
 *   OpenGL ES for the FOSS thinkpad users who can't run anything even remotely
 * modern
 */
namespace fggl::gfx {

	using data::Mesh2D;
	using data::Vertex2D;

	OpenGL4Backend::OpenGL4Backend(const Window &owner) : fggl::gfx::Graphics(), m_canvasPipeline(INVALID_SHADER_ID) {
		// initialise GLEW, or fail
		// FIXME this binds the graphics stack to GLFW :'(
		int version = gladLoadGLLoader( (GLADloadproc)glfwGetProcAddress );
		if (version == 0) {
			printf("Failed to initialize OpenGL context\n");
		}

		// OpenGL debug Support
		GLint flags = 0;
		glGetIntegerv(GL_CONTEXT_FLAGS, &flags);
		if ( (flags & GL_CONTEXT_FLAG_DEBUG_BIT) == GL_CONTEXT_FLAG_DEBUG_BIT) { // NOLINT(hicpp-signed-bitwise)
			debug::info("enabling OpenGL debug output");
			glEnable(GL_DEBUG_OUTPUT);
			glEnable(GL_DEBUG_OUTPUT_SYNCHRONOUS);
			glDebugMessageCallback(fggl_ogl_to_spdlog, nullptr);
			glDebugMessageControl(GL_DONT_CARE, GL_DONT_CARE, GL_DONT_CARE, 0, nullptr, GL_TRUE);
		}

		// setup the viewport based on the window's framebuffer
		auto fbSize = owner.frameSize();
		glViewport(0, 0, fbSize.x, fbSize.y);

		// setup the shader cache
		auto &locator = util::ServiceLocator::instance();
		auto storage = locator.get<data::Storage>();
		m_cache = std::make_unique<ShaderCache>(storage);

		// setup 2D rendering system
		ShaderConfig shader2DConfig = ShaderFromName("canvas");
		m_canvasPipeline = m_cache->load(shader2DConfig);
		if ( m_canvasPipeline == INVALID_SHADER_ID) {
			debug::error("failed to load shader2D - using fallback");
			m_canvasPipeline = m_cache->get(ogl4::FALLBACK_CANVAS_PIPELINE);
		}

		ShaderConfig shader3DConfig = ShaderFromName("phong");
		m_cache->load(shader3DConfig);

		// rendering helpers
		m_canvasRenderer = std::make_unique<ogl4::CanvasRenderer>();
		m_modelRenderer = std::make_unique<ogl4::StaticModelRenderer>(m_cache.get());

		m_debugRenderer = std::make_unique<ogl4::DebugRenderer>(m_cache->getOrLoad(ShaderFromName("debug")));
		if ( m_debugRenderer ) {
			dd::initialize(m_debugRenderer.get());
		}
	};

	void OpenGL4Backend::clear() {
		glClearColor(0.0F, 0.0F, 0.0F, 0.0F);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	}

	void OpenGL4Backend::draw2D(const gfx::Paint &paint) {
		if ( !m_canvasRenderer ) {
			return;
		}

		m_canvasRenderer->render(m_canvasPipeline, paint);
	}

	void OpenGL4Backend::drawScene(ecs3::World &world) {
		if ( m_modelRenderer ) {
			m_modelRenderer->render(world);
		}

		if ( m_debugRenderer ) {
			auto cameras = world.findMatching<gfx::Camera>();
			auto cameraEnt = cameras[0];

			auto* const camTransform = world.get<math::Transform>(cameraEnt);
			auto* const camComp = world.get<gfx::Camera>(cameraEnt);

			const math::mat4 projectionMatrix = glm::perspective(camComp->fov, camComp->aspectRatio, camComp->nearPlane, camComp->farPlane);
			const math::mat4 viewMatrix = glm::lookAt( camTransform->origin(), camComp->target, camTransform->up() );
			m_debugRenderer->mvpMatrix = projectionMatrix * viewMatrix;

			dd::flush();
		}
	}

	void OpenGL4Backend::resize(int width, int height) {
		glViewport(0, 0, width, height);
	}

	// TODO(webpigeon): this shouldn't be hard-coded
	constexpr glm::vec3 DEFAULT_LIGHTPOS = glm::vec3(20.0F, 20.0F, 15.0F);

	void GlMeshRenderer::render(fggl::ecs3::World &ecs, ecs3::entity_t camera, float dt) {
		if (camera == ecs::NULL_ENTITY) {
			debug::warning("tried to render a scene, but no camera exists!");
			return;
		}

		auto entities = ecs.findMatching<GlRenderToken>();
		if (entities.empty()) {
			debug::warning("asked to render, but no entities are renderable");
			return;
		}

		total += dt;

		glEnable(GL_CULL_FACE);
		glCullFace(GL_BACK);

		glEnable(GL_DEPTH_TEST);

		// camera logic
		auto *const camTransform = ecs.get<math::Transform>(camera);
		auto *const camComp = ecs.get<gfx::Camera>(camera);
		glm::mat4 proj = glm::perspective(camComp->fov, camComp->aspectRatio,
										  camComp->nearPlane, camComp->farPlane);
		glm::mat4 view =
			glm::lookAt(camTransform->origin(), camComp->target, camTransform->up());

		// lighting
		glm::vec3 lightPos = DEFAULT_LIGHTPOS;

		// TODO(webpigeon): better performance if grouped by vao first
		// TODO(webpigeon): the nvidia performance presentation said I shouldn't use uniforms for large data
		for (auto &entity : entities) {
			const auto &transform = ecs.get<fggl::math::Transform>(entity);
			const auto &mesh = ecs.get<GlRenderToken>(entity);

			glm::mat4 model = transform->model();
			//	    model = glm::rotate(model, glm::radians(total/2048.0f * 360.0f),
			//glm::vec3(0.0f,1.0f,0.0f));

			auto shader = mesh->pipeline;
			glUseProgram(shader);

			glUniformMatrix4fv(glGetUniformLocation(shader, "model"), 1, GL_FALSE,
							   glm::value_ptr(model));
			glUniformMatrix4fv(glGetUniformLocation(shader, "view"), 1, GL_FALSE,
							   glm::value_ptr(view));
			glUniformMatrix4fv(glGetUniformLocation(shader, "projection"), 1, GL_FALSE,
							   glm::value_ptr(proj));

			// lighting
			GLint lightID = glGetUniformLocation(shader, "lightPos");
			if (lightID != -1) {
				glUniform3fv(lightID, 1, glm::value_ptr(lightPos));
			}

			glBindVertexArray(mesh->vao);

			if (mesh->renderType == GlRenderType::triangle_strip) {
				glEnable(GL_PRIMITIVE_RESTART);
				glPrimitiveRestartIndex(mesh->restartVertex);
			}
			glDrawElements(mesh->renderType, mesh->idxSize, GL_UNSIGNED_INT,
						   reinterpret_cast<void *>(mesh->idxOffset));
			if (mesh->renderType == GlRenderType::triangle_strip) {
				glDisable(GL_PRIMITIVE_RESTART);
			}
		}

		glBindVertexArray(0);
	}

}  // namespace fggl::gfx
