/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef FGGL_ECS_COMPONENT_HPP
#define FGGL_ECS_COMPONENT_HPP

#include "utility.hpp"
#include "fggl/debug/logging.hpp"

#include <cassert>
#include <cstring>
#include <string>
#include <new>
#include <utility>
#include <iostream>

#include <vector>
#include <unordered_map>
#include "yaml-cpp/yaml.h"

namespace fggl::ecs {

	template<typename T>
	bool restore_config(T* comp, const YAML::Node& node) {
		return false;
	}

	class ComponentBase {
		public:
			using data_t = unsigned char;

			virtual ~ComponentBase() {};

			// in place
			virtual void destroy(data_t *data) const = 0;
			virtual void move(data_t *src, data_t *dest) const = 0;
			virtual void bulkMove(data_t *src, data_t *dest, std::size_t count) = 0;
			virtual void construct(data_t *data) const = 0;

			// virtual
			virtual void *construct() const = 0;
			virtual void *restore(const YAML::Node& config) const = 0;

			virtual void *copyConstruct(const void *src) = 0;

			virtual const char *name() const = 0;
			virtual const component_type_t id() const = 0;

			virtual std::size_t size() const = 0;
	};

	template<class C>
	class Component : public ComponentBase {
		public:
			virtual void destroy(data_t *data) const override {
				C *location = std::launder(reinterpret_cast<C *>(data));
				location->~C();
			}

			virtual const char *name() const override {
				return C::name;
			}

			virtual const component_type_t id() const {
				return Component<C>::typeID();
			}

			virtual void construct(unsigned char *data) const override {
				new(data) C();
			}

			virtual void* restore(const YAML::Node& config) const override {
				C* ptr = new C();
				bool restored = restore_config<C>(ptr, config);
				if ( !restored ) {
					debug::warning("error restoring {}", C::name);
					assert( false && "failed to restore configuration when loading type!" );
				}
				return ptr;
			}

			void *copyConstruct(const void *src) override {
				const C *srcPtr = (C *) src;
				return new C(*srcPtr);
			}

			void *construct() const override {
				return new C();
			}

			virtual void move(data_t *src, data_t *dest) const override {
				assert(src != nullptr);
				assert(dest != nullptr);
				new(&dest[0]) C(std::move(*reinterpret_cast<C *>(src)));
			}

			virtual void bulkMove(data_t *src, data_t *dest, std::size_t count) {
				if (std::is_trivially_copyable<C>::value) {
					std::memcpy(dest, src, count * size());
				} else {
					unsigned char *srcPtr = src;
					unsigned char *destPtr = dest;

					for (std::size_t i = 0; i < count; ++i) {
						new(destPtr) C(std::move(*reinterpret_cast<C *>(srcPtr)));
						srcPtr += sizeof(C);
						destPtr += sizeof(C);
					}
				}
			}

			virtual std::size_t size() const {
				return sizeof(C);
			}

			static component_type_t typeID() {
				return TypeIdGenerator<ComponentBase>::GetNewID<C>();
			}
	};

}

#endif
