/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 23/10/2021.
//

#ifndef FGGL_ECS3_FAST_CONTAINER_HPP
#define FGGL_ECS3_FAST_CONTAINER_HPP

#include <cstdint>
#include <cstdarg>
#include <cassert>

#include <fggl/ecs3/types.hpp>

namespace fggl::ecs3 {


	class Container {
		public:
			const RecordIdentifier m_identifier;

			Container(const TypeRegistry &reg, RecordIdentifier id) :
                m_identifier(id),
				m_types(reg),
				backingStore(nullptr),
                m_size(0),
                m_capacity(0)  {};

			~Container() {
				delete[] backingStore;
			}

			std::size_t create();

			void remove(std::size_t pos);

			std::size_t expand(Container &other, std::size_t otherPos, component_type_t newComp);

			void contract(Container &other, std::size_t otherPos);

			void ensure(std::size_t size);

			inline unsigned char *data_raw(component_type_t type) {
				auto seek_id = m_identifier.idx(type);

				// you asked for something I don't contain...
				if (seek_id == m_identifier.count) {
					std::cerr << "asked for " << type << " from " << m_identifier << std::endl;
					assert(seek_id != m_identifier.count);
					return nullptr;
				}

				// figure out the offset
				return backingStore + offsets[seek_id];
			}

			template<typename T>
			inline T *data() {
				auto comp_id = Component<T>::typeID();
				return (T *) data_raw(comp_id);
			}

			template<typename T>
			T *set(std::size_t entity, T *compData) {
				auto *comps = data<T>();
				auto entityPos = idx(entity);

				auto compMeta = m_types.template meta<T>();

				unsigned char *usrPtr = (unsigned char *) &comps[entityPos];
				compMeta->destroy(usrPtr);
				compMeta->move((unsigned char *) compData, usrPtr);

				return &comps[entityPos];
			}

			[[nodiscard]]
			inline std::size_t size() const {
				return m_size;
			}

			inline std::size_t idx(entity_t entity) {
				auto *entityData = data<EntityMeta>();
				for (std::size_t i = 0; i < m_size; i++) {
					if (entityData[i].id == entity) {
						return i;
					}
				}
				return m_size;
			}

		private:
			const TypeRegistry &m_types;
			unsigned char *backingStore;
			std::size_t offsets[RecordIdentifier::MAX_COMPS]{};
			std::size_t m_size;
			std::size_t m_capacity;

			void move(std::size_t newPos, Container &oldContainer, std::size_t oldPos);
	};

}

#endif //FGGL_ECS3_FAST_CONTAINER_HPP
