/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef FGGL_ECS3_FAST_ECS_HPP
#define FGGL_ECS3_FAST_ECS_HPP

#include <cstddef>
#include <cstdarg>
#include <cassert>
#include <cstring>

#include <map>
#include <algorithm>
#include <iostream>
#include <type_traits>

#include <fggl/ecs3/utils.hpp>
#include <fggl/ecs3/types.hpp>
#include <fggl/ecs3/fast/Container.hpp>

namespace fggl::ecs3::fast {

	using entity_t = unsigned int;
	constexpr entity_t NULL_ENTITY = 0;

	class World {
		public:
			explicit World(TypeRegistry &reg) : m_registry(reg), m_last(NULL_ENTITY) {}

			entity_t create() {
				auto next = m_last++;

				auto arch = make_id(1, Component<EntityMeta>::typeID());
				auto &container = getContainer(arch);

				m_entities[next] = container.m_identifier;

				auto pos = container.create();
				auto *entityMeta = container.data<EntityMeta>();
				entityMeta[pos].id = next;

				return next;
			}

			void remove(entity_t entity) {
				auto arch = m_entities.at(entity);
				auto container = m_records.at(arch);

				auto entPos = container.idx(entity);
				container.remove(entPos);
			}

			inline Container &getContainer(RecordIdentifier &arch) {
				try {
					return m_records.at(arch);
				} catch (std::out_of_range &e) {
					auto v = m_records.emplace(std::pair<RecordIdentifier, Container>(arch, {m_registry, arch}));
					return v.first->second;
				}
			}

			template<typename T>
			T *add(const entity_t entity) {
				auto currArch = m_entities.at(entity);
				auto newArch = currArch.with<T>();
				m_entities[entity] = newArch;

				auto &oldContainer = m_records.at(currArch);
				auto &newContainer = getContainer(newArch);

				auto oldPos = oldContainer.idx(entity);

				auto newPos = newContainer.expand(oldContainer, oldPos, Component<T>::typeID());
				auto *data = newContainer.template data<T>();
				return &data[newPos];
			}

			template<typename T>
			T *set(const entity_t entity, T *record) {
				auto currArch = m_entities.at(entity);

				// check we already have that component type...
				if (currArch.idx(Component<T>::typeID()) == currArch.count) {
					add<T>(entity);
					currArch = m_entities.at(entity);
				}

				auto &container = m_records.at(currArch);
				auto pos = container.idx(entity);
				return container.set<T>(pos, record);
			}

			template<typename T>
			T *get(entity_t entity) {
				auto currArch = m_entities.at(entity);
				auto pos = currArch.idx(entity);
				auto &container = m_records.at(currArch);

				auto *data = container.template data<T>();
				return &data[pos];
			}

			template<typename T>
			const T *get(entity_t entity) const {
				auto currArch = m_entities.at(entity);
				auto pos = currArch.idx(entity);
				auto &container = m_records.at(currArch);

				auto *data = container.template data<T>();
				return &data[pos];
			}

			template<typename T>
			void remove(entity_t entity) {
				auto currArch = m_entities.at(entity);
				auto newArch = currArch.without<T>();

				auto &oldContainer = m_records[currArch];
				auto &newContainer = m_records[newArch];

				auto oldPos = oldContainer.idx(entity);
				auto newPos = newContainer.create();

				m_records[newArch].contract(newPos, oldContainer, oldPos);
			}

			template<typename... T>
			std::vector<entity_t> findMatching() const {
				return {};
			}

		private:
			TypeRegistry &m_registry;
			std::map<RecordIdentifier, Container> m_records;
			std::map<entity_t, RecordIdentifier> m_entities;
			entity_t m_last{};
	};

}

#endif
