/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 23/10/2021.
//

#ifndef FGGL_ECS3_MODULE_MODULE_HPP
#define FGGL_ECS3_MODULE_MODULE_HPP

#include <string>
#include <map>
#include <memory>

#include <fggl/ecs3/types.hpp>

namespace fggl::ecs3 {

	class Module {
		public:
			virtual ~Module() = default;

			[[nodiscard]] virtual std::string name() const = 0;

			virtual void onLoad(ModuleManager &manager, TypeRegistry &tr) {};

			virtual void onUpdate();
			virtual void onFrameStart();
			virtual void onFrameEnd();
	};

	class ModuleManager {
		public:
			explicit ModuleManager(TypeRegistry &types) : m_types(types), m_modules() {}

			~ModuleManager() = default;

			template<typename C, typename... Args>
			std::shared_ptr<C> load(Args &... args) {
				auto ptr = std::make_shared<C>(args...);
				m_modules[ptr->name()] = ptr;

				ptr->onLoad(*this, m_types);
				//spdlog::info("loaded ECS module: {}", ptr->name());
				return ptr;
			}

			template<typename C>
			void onAdd(const callback_t &cb) {
				m_types.callbackAdd(Component<C>::typeID(), cb);
			}

			void onUpdate() {
				for (auto &[id, ptr] : m_modules) {
					ptr->onUpdate();
				}
			}

			void onFrameStart() {
				for (auto &[id, ptr] : m_modules) {
					ptr->onFrameStart();
				}
			}

			void onFrameEnd() {
				for (auto &[id, ptr] : m_modules) {
					ptr->onFrameEnd();
				}
			}

		private:
			TypeRegistry &m_types;
			std::map<std::string, std::shared_ptr<Module>> m_modules;
	};

}

#endif //FGGL_ECS3_MODULE_MODULE_HPP
