/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 04/06/22.
//

#ifndef FGGL_ECS3_PROTOTYPE_LOADER_HPP
#define FGGL_ECS3_PROTOTYPE_LOADER_HPP

#include "yaml-cpp/yaml.h"

#include "fggl/math/types.hpp"
#include "fggl/phys/types.hpp"
#include "fggl/gfx/phong.hpp"

#include "fggl/data/storage.hpp"
#include "fggl/data/model.hpp"
#include "fggl/data/procedural.hpp"

#include "fggl/ecs3/ecs.hpp"

namespace fggl::ecs3 {

	void load_prototype_node( ecs3::World& world, const YAML::Node& node);
	void load_prototype_file( ecs3::World& world, data::Storage& storage, const std::string& name );

} // namespace fggl::ecs3

namespace YAML {

	template<>
	struct convert<fggl::math::vec3> {
		static Node encode(const fggl::math::vec3& rhs){
			Node node;
			node.push_back(rhs.x);
			node.push_back(rhs.y);
			node.push_back(rhs.z);
			return node;
		}

		static bool decode(const Node& node, fggl::math::vec3& rhs) {
			if ( !node.IsSequence() || node.size() != 3) {
				return false;
			}

			rhs.x = node[0].as<float>();
			rhs.y = node[1].as<float>();
			rhs.z = node[2].as<float>();
			return true;
		}
	};

	template<>
	struct convert<fggl::math::quat> {
		static Node encode(const fggl::math::quat& rhs){
			Node node;
			node.push_back(rhs.x);
			node.push_back(rhs.y);
			node.push_back(rhs.z);
			node.push_back(rhs.w);
			return node;
		}

		static bool decode(const Node& node, fggl::math::quat& rhs) {
			if ( !node.IsSequence() || node.size() != 3) {
				return false;
			}

			rhs.x = node[0].as<float>();
			rhs.y = node[1].as<float>();
			rhs.z = node[2].as<float>();
			rhs.w = node[3].as<float>();
			return true;
		}
	};

	constexpr const char* TYPE_KINEMATIC = "kinematic";
	constexpr const char* TYPE_STATIC = "static";
	constexpr const char* TYPE_DYNAMIC = "dynamic";

	template<>
	struct convert<fggl::phys::BodyType> {
		static Node encode(const fggl::phys::BodyType& rhs) {
			switch (rhs) {
				case fggl::phys::BodyType::KINEMATIC:
					return Node(TYPE_KINEMATIC);
				case fggl::phys::BodyType::STATIC:
					return Node(TYPE_STATIC);
				default:
				case fggl::phys::BodyType::DYNAMIC:
					return Node(TYPE_DYNAMIC);
			}
		}

		static bool decode(const Node& node, fggl::phys::BodyType& rhs) {
			const auto value = node.as<std::string>();
			if ( value == TYPE_KINEMATIC ) {
				rhs = fggl::phys::BodyType::KINEMATIC;
			} else if ( value == TYPE_STATIC ) {
				rhs = fggl::phys::BodyType::STATIC;
			} else {
				rhs = fggl::phys::BodyType::DYNAMIC;
			}
			return true;
		}
	};


}

namespace fggl::ecs {

	constexpr int DEFAULT_STACKS = 16;
	constexpr int DEFAULT_SLICES = 16;
	constexpr const char* SHAPE_SPHERE_VALUE{"sphere"};
	constexpr const char* SHAPE_BOX_VALUE{"box"};

	// scene template specialisations
	template<>
	bool restore_config(math::Transform* comp, const YAML::Node& node);

	template<>
	bool restore_config(gfx::PhongMaterial* comp, const YAML::Node& node);

	template<>
	bool restore_config(data::StaticMesh* meshComp, const YAML::Node& node);

	template<>
	bool restore_config(phys::RigidBody* body, const YAML::Node& node);

	template<>
	bool restore_config(phys::CollisionCallbacks* callbacks, const YAML::Node& node);

	template<>
	bool restore_config(phys::CollisionCache* cache, const YAML::Node& node);
}

#endif //FGGL_ECS3_PROTOTYPE_LOADER_HPP
