/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 23/10/2021.
//

#ifndef FGGL_ECS3_PROTOTYPE_WORLD_HPP
#define FGGL_ECS3_PROTOTYPE_WORLD_HPP

#include <map>
#include <functional>
#include <unordered_set>

#include "fggl/ecs3/types.hpp"
#include "fggl/debug/logging.hpp"
#include <yaml-cpp/yaml.h>

/**
 * A component based implementation of a game world.
 *
 * This is not a true ECS but exposes a similar API to it for testing (with a lot less headaches).
 */
namespace fggl::ecs3::prototype {

	using EntityCallback = std::function<void(const entity_t)>;

	class Entity {
		public:
			bool m_abstract;

			explicit Entity(entity_t id) : m_abstract(false), m_id(id) {};

			Entity(const Entity &entity) : m_id(entity.m_id), m_components(entity.m_components) {
				//spdlog::info("entity created fro copy: {}", m_id);
			}

			~Entity() = default;

			template<typename C>
			C *add() {
				C *ptr = new C();
				m_components[Component<C>::typeID()] = ptr;
				return ptr;
			}

			void *add(std::shared_ptr<ComponentBase> t) {
				void *ptr = t->construct();
				m_components[t->id()] = ptr;
				return ptr;
			}

			void* add(const std::shared_ptr<ComponentBase>& compMeta, const YAML::Node& config) {
				void* ptr = compMeta->restore(config);
				m_components[ compMeta->id() ] = ptr;
				return ptr;
			}

			template<typename C>
			C *set(const C *ptr) {
				C *newPtr = new C(*ptr);
				m_components[Component<C>::typeID()] = newPtr;
				return newPtr;
			}

			void *set(const std::shared_ptr<ComponentBase> &t, const void *ptr) {
				void *newPtr = t->copyConstruct(ptr);
				m_components[t->id()] = newPtr;
				return newPtr;
			}

			template<typename C>
			C *get() const {
				void *ptr = m_components.at(Component<C>::typeID());
				return (C *) ptr;
			}

			template<typename C>
			void remove(){
				m_components.erase(Component<C>::typeID());
			}

			inline void *get(component_type_t t) {
				return m_components.at(t);
			}

			std::vector<component_type_t> getComponentIDs() {
				std::vector<component_type_t> comps{};
				for (auto &[k, _] : m_components) {
					comps.push_back(k);
				}
				return comps;
			}

			bool hasComponents(std::vector<component_type_t> &Cs) const {
				for (auto c : Cs) {
					if (m_components.find(c) == m_components.end()) {
						return false;
					}
				}
				return true;
			}

		private:
			entity_t m_id;
			std::map<component_type_t, void *> m_components;
	};

	class World {
		public:
			explicit World(TypeRegistry &reg) : m_types(reg), m_next(1), m_entities() {};
			~World() = default;

			entity_t create(bool abstract) {
				auto nextID = m_next++;
				m_entities.emplace(nextID, nextID);

				auto &entity = m_entities.at(nextID);
				entity.m_abstract = abstract;

				// meta data
				auto *meta = entity.add<ecs3::EntityMeta>();
				meta->id = nextID;
				meta->abstract = abstract;
				meta->typeName = "";

				return nextID;
			}

			inline entity_t createFromPrototype(const std::string& name) {
				return copy(findPrototype(name) );
			}

			entity_t copy(entity_t prototype) {
				auto clone = create(false);

				auto components = getComponents(prototype);
				for (auto component : components) {
					auto protoComp = get(prototype, component);
					set(clone, component, protoComp);
				}

				return clone;
			}

			void addFromConfig(entity_t entity, component_type_t type, const YAML::Node& node) {
				auto meta = m_types.meta(type);

				auto& entityObj = m_entities.at(entity);
				entityObj.add(meta, node);

				m_types.fireAdd(this, entity, meta->id());
			}

			void createFromSpec(entity_t entity, const YAML::Node& compConfig) {
				if ( compConfig ) {
					for (const auto& itr : compConfig ) {
						const auto name = itr.first.as<std::string>();
						const auto& config = itr.second;

						auto compType = m_types.find( name.c_str() );
						addFromConfig(entity, compType, config);
					}
				}
			}

			inline auto alive(entity_t entity) const -> bool {
				return entity != NULL_ENTITY
					&& m_killList.find( entity ) == m_killList.end()
					&& m_entities.find( entity ) != m_entities.end();
			}

			inline auto exists(entity_t entity) const -> bool {
				return entity != NULL_ENTITY
					&& m_entities.find( entity ) != m_entities.end();
			}

			void destroy(entity_t entity) {
				assert( alive(entity) && "attempted to kill null entity" );
				// TOOD resolve and clean components
				//m_entities.erase(entity);
				m_killList.insert(entity);
			}

			void reapEntities() {
				for (const auto& entity : m_killList) {
					//auto& entityObj = m_entities.at(entity);
					//entityObj.clear();
					for (auto& listener : m_deathListeners) {
						listener( entity );
					}

					m_entities.erase(entity);
				}
				m_killList.clear();
			}

			inline TypeRegistry &types() {
				return m_types;
			}

			std::vector<entity_t> all() {
				std::vector<entity_t> entities{};
				for (auto &[eid, entity] : m_entities) {
					entities.push_back(eid);
				}
				return entities;
			}

			std::vector<component_type_t> getComponents(entity_t entityID) {
				assert(alive(entityID) && "attempted to get components on dead entity");

				std::vector<component_type_t> components{};
				auto &entity = m_entities.at(entityID);
				auto comps = entity.getComponentIDs();
				for (auto id : comps) {
					components.push_back(id);
				}
				return components;
			}

			template<typename... Cs>
			std::vector<entity_t> findMatching() const {
				// construct the key
				std::vector<ecs::component_type_t> key;
				(key.push_back(Component<Cs>::typeID()), ...);

				// entities
				std::vector<entity_t> entities{};
				for (auto &[eid, entity] : m_entities) {
					if (entity.hasComponents(key) && !entity.m_abstract) {
						entities.push_back(eid);
					}
				}

				return entities;
			}

			template<typename ...Cs>
			bool has(entity_t entityIdx) const {
				if ( !alive(entityIdx)) {
					return false;
				}

				std::vector<ecs::component_type_t> key;
				(key.push_back(Component<Cs>::typeID()), ...);
				return m_entities.at(entityIdx).hasComponents(key);
			}

			template<typename C>
			C *add(entity_t entity_id) {
				assert( alive(entity_id) && "attempted to add component on null entity" );

				//spdlog::info("component '{}' added to '{}'", C::name, entity_id);
				auto &entity = m_entities.at(entity_id);
				auto comp = entity.template add<C>();

				m_types.fireAdd(this, entity_id, Component<C>::typeID());
				return comp;
			}

			void *add(entity_t entity_id, component_type_t component_id) {
				assert( alive(entity_id) && "attempted to add component on null entity" );

				auto meta = m_types.meta(component_id);
				auto &entity = m_entities.at(entity_id);

				void *ptr = entity.add(meta);
				m_types.fireAdd(this, entity_id, meta->id());
				return ptr;
			}

			template<typename C>
			C *set(entity_t entity_id, const C *ptr) {
				assert( alive( entity_id ) && "attempted to set component on null entity" );

				//spdlog::info("component '{}' set on '{}'", C::name, entity_id);
				auto &entity = m_entities.at(entity_id);
				auto comp = entity.set<C>(ptr);

				m_types.fireAdd(this, entity_id, Component<C>::typeID());
				return comp;
			}

			void *set(entity_t entity_id, component_type_t cid, const void *ptr) {
				assert( alive( entity_id ) && "attempted to set component on null entity" );

				auto &entity = m_entities.at(entity_id);
				auto cMeta = m_types.meta(cid);

				auto comp = entity.set(cMeta, ptr);
				m_types.fireAdd(this, entity_id, cid);
				return comp;
			}

			template<typename C>
			C* tryGet(entity_t entity_id) const {
				if ( entity_id == NULL_ENTITY) {
					return nullptr;
				}

				try {
					return get<C>(entity_id);
				} catch ( std::out_of_range& e) {
					fggl::debug::info("component {} on {} did not exist", C::name, entity_id);
					return nullptr;
				}
			}

			template<typename C>
			C *get(entity_t entity_id) const {
				assert( exists(entity_id) && "attempted to get component on null entity" );

				const auto& entity = m_entities.at(entity_id);
				return entity.get<C>();
			}

			template<typename C>
			void remove(entity_t entity_id) {
				assert( alive(entity_id) && "attempted to remove component on null entity" );

				try {
					auto &entity = m_entities.at(entity_id);
					try {
						return entity.remove<C>();
					} catch ( std::out_of_range& e ) {
						std::cerr << "entity " << entity_id << " does not have component "<< C::name << std::endl;
					}
				} catch ( std::out_of_range& e) {
					std::cerr << "tried to delete component on entity that didn't exist, entity was: " << entity_id << std::endl;
				}
			}

			void *get(entity_t entity_id, component_type_t componentType) {
				assert( exists(entity_id) && "attempted to get component on null entity" );
				auto &entity = m_entities.at(entity_id);
				return entity.get(componentType);
			}

			void addDeathListener(const EntityCallback& callback) {
				m_deathListeners.emplace_back(callback);
			}

			entity_t findPrototype(const std::string& name) const {
				for ( const auto& [entity, obj] : m_entities ) {
					if ( !obj.m_abstract ){
						continue;
					}

					auto* metaData = obj.get<EntityMeta>();
					if ( metaData->typeName == name){
						return entity;
					}
				}

				return NULL_ENTITY;
			}

		private:
			std::vector<EntityCallback > m_deathListeners;
			TypeRegistry &m_types;
			entity_t m_next;
			std::map<entity_t, Entity> m_entities;
			std::unordered_set<entity_t> m_killList;

	};

}

#endif //FGGL_ECS3_PROTOTYPE_WORLD_HPP
