/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef FGGL_ECS3_TYPES_HPP
#define FGGL_ECS3_TYPES_HPP

#include <cstdarg>
#include <utility>
#include <functional>

#include <fggl/ecs/component.hpp>
#include <fggl/ecs3/utils.hpp>

#include <iostream>
#include <memory>
#include <algorithm>
#include <map>
#include <unordered_map>

namespace fggl::ecs3 {

	namespace {
		using namespace fggl::ecs;
	};

	namespace prototype {
		class World;
	}

	using fggl::ecs::component_type_t;

	class ModuleManager;

	using callback_t = std::function<void(prototype::World *, ecs3::entity_t)>;
	struct TypeCallbacks {
		std::vector<callback_t> add;
	};

	// core component types
	struct EntityMeta {
		constexpr static const char name[] = "meta";
		entity_t id;
		bool abstract;
		std::string typeName;
	};

	struct RecordIdentifier {
		constexpr static std::size_t MAX_COMPS = 32;
		component_type_t types[MAX_COMPS];
		std::size_t count;

		[[nodiscard]]
		inline std::size_t idx(component_type_t t) const {
			return utils::search(types, count, t);
		}

		template<typename T>
		[[nodiscard]]
		RecordIdentifier with() const {
			// check the caller wasn't a muppet
			const auto typeID = ecs::Component<T>::typeID();
			if (idx(typeID) != count) {
				return *this;
			}

			RecordIdentifier re{};
			re.count = count + 1;
			re.types[count] = ecs::Component<T>::typeID();

			// add old types
			for (std::size_t i = 0; i < count; ++i) {
				re.types[i] = types[i];
			}
			std::sort(re.types, re.types + re.count);
			return re;
		}

		template<typename T>
		[[nodiscard]]
		RecordIdentifier without() const {
			// check the caller wasn't a muppet
			const auto typeID = ecs::Component<T>::typeID();
			const auto typeIdx = idx(typeID);
			if (typeIdx == count) {
				return *this;
			}

			RecordIdentifier re{};
			re.count = count - 1;

			// add old types
			for (std::size_t i = 0, j = 0; i < count; ++i) {
				if (typeIdx != i) {
					re.types[j] = types[i];
					j++;
				}
			}
			std::sort(re.types, re.types + re.count);
			return re;
		}

		bool operator<(const RecordIdentifier &other) const {
			if (count < other.count) {
				return true;
			} else if (count > other.count) {
				return false;
			} else {
				for (std::size_t i = 0; i < count; i++) {
					if (types[i] != other.types[i]) {
						return types[i] < other.types[i];
					}
				}
				return false;
			}
		}

		bool operator==(const RecordIdentifier &arg) const {
			if (arg.count != count) {
				return false;
			}

			for (std::size_t i = 0; i < count; i++) {
				if (types[i] != arg.types[i]) {
					return false;
				}
			}
			return true;
		}

		bool operator!=(const RecordIdentifier &arg) const {
			return !(*this == arg);
		}
	};

	std::ostream &operator<<(std::ostream &out, RecordIdentifier const &curr);

	inline RecordIdentifier make_id(std::size_t count, ...) {
		assert(count < RecordIdentifier::MAX_COMPS);

		RecordIdentifier re{};

		std::va_list args;
		va_start(args, count);
		for (std::size_t i = 0; i < count; ++i) {
			re.types[i] = va_arg(args, component_type_t);
		}
		va_end(args);
		re.count = count;
		std::sort(re.types, re.types + count);

		return re;
	}

	class TypeRegistry {
		public:

			TypeRegistry() : m_last_virtual(9000), m_callbacks() {
				// core types always exist
				make<EntityMeta>();
			}

			template<typename T>
			void make() {
				auto type_id = Component<T>::typeID();
				if (m_types.find(type_id) != m_types.end())
					return;
				m_types[type_id] = std::make_shared<Component<T>>();
			}

			template<typename T>
			bool exists() {
				auto type_id = Component<T>::typeID();
				return m_types.find(type_id) != m_types.end();
			}

			template<typename T>
			std::shared_ptr<fggl::ecs::ComponentBase> meta() const {
				auto type_id = Component<T>::typeID();
				return m_types.at(type_id);
			}

			inline std::shared_ptr<fggl::ecs::ComponentBase> meta(component_type_t type_id) const {
				try {
					return m_types.at(type_id);
				} catch (std::out_of_range& err) {
					std::cerr << "asked for metadata on type " << type_id << " but no such type is in the type system" << std::endl;
					return nullptr;
				}
			}

			inline component_type_t find(const char *name) const {
				for (const auto &[type, meta] : m_types) {
					if (std::strcmp(name, meta->name()) == 0) {
						return type;
					}
				}

				debug::warning("asked for unknown/unregistered component type: {}", name);
				assert(false && "unknown component type, are you sure it was registered?");
				return 0;
			}

			inline void make_virtual(std::shared_ptr<ComponentBase> vtype) {
				auto type_id = m_last_virtual++;
				m_types[type_id] = std::move(vtype);
			}

			void callbackAdd(component_type_t component, const callback_t &callback) {
				m_callbacks[component].add.push_back(callback);
			}

			void fireAdd(prototype::World *world, entity_t entity, component_type_t type) {
				try {
					auto &callbacks = m_callbacks.at(type).add;
					for (auto &callback : callbacks) {
						callback(world, entity);
					}
				} catch (std::out_of_range &e) {
					//spdlog::debug("no callbacks for {}", m_types[type]->name());
				}
			}

		private:
			std::unordered_map<component_type_t, std::shared_ptr<fggl::ecs::ComponentBase>> m_types;
			component_type_t m_last_virtual;
			std::map<component_type_t, TypeCallbacks> m_callbacks;
	};

};

#endif
